import base64
from os import path
from time import time_ns
from urllib.parse import quote, urlparse


def masking_string(value: str, start=1, count=1):
    """
    문자열 일부를 "*" 값으로 마스킹 처리한 값을 반환
    :param value: 대상 문자열
    :param start: 시작위치
    :param count: 변환할 문자 갯수
    :return: 변환된 문자
    """
    target_len = value.__len__()

    if target_len - start < count:
        raise ValueError("Value must be bigger than count")
    elif count == 0:
        raise ValueError("Count must be bigger than zero")

    mask = "*" * count
    return value[:start] + mask + value[start + count :]


def string_to_base64(target, to_string=True):
    """
    문자열을 base64 형식으로 변환
    :param target: 변환대상
    :param to_string: 문자열 반환여부, False인 경우 bytes 반환
    """
    code = base64.b64encode(target.encode("utf-8"))
    if to_string:
        return code.decode("utf-8")
    return code


def base64_to_string(target: str):
    """
    base64 문자열을 decode 한 문자열 반환
    """
    return base64.b64decode(target).decode("utf-8")


def number_with_commas(number: int) -> str:
    return "{:,}".format(number)


def url_encode(value: str, encode="utf-8"):
    return quote(value, encode)


def is_mobile_number(value: str) -> bool:
    """
    휴대폰 번호 여부 체크
    """
    return value and value[:2] == "01"


def parse_tel(value) -> str:
    """
    전화번호 문자열에서 "-" 제거한 문자열 반환
    """
    raw = "" if type(value) != str else value
    return raw.replace("-", "")


def snake_to_camel(value: str) -> str:
    combined = value.split("_")
    return combined[0] + "".join(x.title() for x in combined[1:])


def get_nano_string(div=1):
    """
    nano timestamp to string
    """
    return str(time_ns() // div)


def remove_space(value: str):
    """
    문자열의 공백제거
    """
    return "".join(value.split())



from typing import Union


def round_by_position(value: Union[int, float], position=-1):
    """
    자릿수 기준 반올림
    기본값은 10자리 단위 반올림
    ex. 22 -> 20, 255 -> 260, 254 -> 250
    """
    return int(round(float(value), position))



from dataclasses import dataclass
from enum import Enum
from typing import Optional


class SlackButtonStyle(str, Enum):

    primary = "primary"
    danger = "danger"


@dataclass
class SlackButton:

    text: str
    link: Optional[str] = None
    style: Optional[SlackButtonStyle] = None



import json
from typing import List, Union

import requests

from .types import SlackButton


class SlackMessageBuilder:
    """
    Slack message builder
    - Builder pattern 으로 구축, 사용
    - 생성 후 add_*** -> build -> send
    """

    def __init__(self, debug_channel: str, active=True, debug=False):
        self.active = active
        self.debug = debug
        self.debug_channel = debug_channel
        self.data = {}
        self.block = []

    @staticmethod
    def transform_text(text: str):
        return {"type": "mrkdwn", "text": text}

    def send(self, hook_url, **kwargs):
        url = hook_url

        if not self.active:
            print("(Slack)Not active")
            return
        elif self.debug:
            url = self.debug_channel

        if not url:
            print(f"(Slack)Not provided url - {self.data}")
            return

        return requests.post(
            url,
            data=self.data,
            headers={"Content-Type": "application/json"},
            **kwargs,
        )

    def build(self):
        self.data = json.dumps({"blocks": self.block})
        return self

    def add_divider(self):
        """
        구분선
        https://api.slack.com/reference/block-kit/blocks#divider
        """
        self.block.append({"type": "divider"})
        return self

    def add_text(self, text: Union[str, List[str]]):
        """
        섹션 형태로 텍스트 구성
        https://api.slack.com/reference/block-kit/blocks#section
        :param text: 문자열 목록 또는 문자열
        """
        if type(text) == list:
            fields = [self.transform_text(o) for o in text]
            self.block.append({"type": "section", "fields": fields})
        else:
            self.block.append({"type": "section", "text": self.transform_text(text)})

        return self

    def add_image(self, url, alt, **kwargs):
        """
        이미지 추가
        https://api.slack.com/reference/block-kit/blocks#image
        :param url: 이미지 URL (~3000)
        :param alt: 이미지 ALT text (~2000)
        :param kwargs: title, block_id (optional)
        """
        self.block.append(
            {
                "type": "image",
                "image_url": url,
                "alt_text": alt,
                **kwargs,
            }
        )
        return self

    def add_header(self, text: str):
        """
        헤더 스타일 텍스트
        :param text: 문자열
        """
        self.block.append(
            {
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": text,
                    "emoji": True,
                },
            }
        )
        return self

    def add_button(self, buttons: List[SlackButton]):
        """
        버튼
        https://api.slack.com/reference/block-kit/block-elements#button
        :return: 버튼 Dataclass
        """
        elements = []

        for button in buttons:
            elem = {
                "type": "button",
                "text": {
                    "type": "plain_text",
                    "emoji": True,
                    "text": button.text,
                },
            }

            if button.link:
                elem["url"] = button.link
            if button.style:
                elem["style"] = button.style.value

            elements.append(elem)

        self.block.append(
            {
                "type": "button",
                "elements": elements,
            }
        )
        return self

    def add_footer(self, text: str):
        """
        푸터 스타일 텍스트
        :param text: 문자열
        """
        self.block.append(
            {
                "type": "context",
                "elements": [
                    {
                        "type": "mrkdwn",
                        "text": text,
                    },
                ],
            },
        )
        return self



from typing import List


def exclude_dict_empty_values(dictionary: dict):
    """
    빈 문자열 또는 None 값을 가진 key를 dict에서 제거
    """
    return {k: v for k, v in dictionary.items() if v != "" and v is not None}


def split_elements_per_unit(values: list, unit: int, remove_none=True):
    """
    주어진 리스트를 unit 갯수에 맞게 잘라 리스트로 삽입
    ie. ['a','b','c','d'] 를 unit 3을 적용 -> [['a','b','c'], ['d', None, None, None]]
    remove 옵션 적용시 None 은 제거됨.
    """
    total_length = len(values)
    extended_list = (
        values + [None] * (unit - (total_length % unit))
        if total_length % unit != 0
        else values
    )
    if remove_none:
        return [
            list(filter(bool, extended_list[i : i + unit]))
            for i in range(0, len(extended_list), unit)
        ]
    return [extended_list[i : i + unit] for i in range(0, len(extended_list), unit)]


def dict_list_to_map(key, values: List[dict]):
    """
    dict list를 dict 요소의 주어진 키에 대한 값을 기준으로 dict map으로 전환
    :param key: dict 내부의 키
    :param values: 대상값
    :return: 변환된 dict
    """
    return {obj[key]: obj for obj in values}


def omit(dictionary: dict, keys: list):
    """
    dict 에서 특정한 키를 제외하고 반환
    :param dictionary: 대상
    :param keys: 제외할 key
    :return: dict
    """
    return {key: dictionary[key] for key in dictionary if key not in keys}


def pick(dictionary: dict, keys: list):
    """
    dict 에서 특정한 키만 선택하여 반환
    :param dictionary: 대상
    :param keys: 선택할 key
    :return: dict
    """
    return {key: dictionary[key] for key in dictionary if key in keys}


def chunk_list(input_list: list, chunk_size: int):
    """
    입력받은 목록을 원하는 최대 사이즈 내에서 조각내어 반환
    :param input_list: 입력 목록
    :param chunk_size: 최대 사이즈
    :return: generator
    """
    for i in range(0, len(input_list), chunk_size):
        yield input_list[i : i + chunk_size]


def get_value(obj, key: str, default=None, raise_exception=False):
    """
    Dict 타입 혹은 instance 를 받아 키에 해당하는 값을 반환
    :param obj: Dict or Any object
    :param key: 탐색할 키
    :param default: 기본값
    :param raise_exception: True 인 경우 키의 값이 없을 때 KeyError or AttributeError raise
    :return: value
    """
    if isinstance(obj, dict):
        return obj[key] if raise_exception else obj.get(key, default)
    return getattr(obj, key) if raise_exception else getattr(obj, key, default)


def extract_terminal_strings(data: dict):
    """
    Dict 타입의 값을 받아 재귀탐색하여 말단의 문자열들을 모아 List로 반환
    :param data: Dict
    :return: List[str]
    """
    terminal_strings = []

    def extract_strings_from_dict(d):
        for key, value in d.items():
            if isinstance(value, dict):
                extract_strings_from_dict(value)
            elif isinstance(value, str):
                terminal_strings.append(value)

    extract_strings_from_dict(data)
    return terminal_strings



from datetime import datetime, timedelta, timezone, tzinfo


def is_aware(value: datetime):
    """
    Determine if a given datetime.datetime is aware.

    The concept is defined in Python's docs:
    https://docs.python.org/library/datetime.html#datetime.tzinfo

    Assuming value.tzinfo is either None or a proper datetime.tzinfo,
    value.utcoffset() implements the appropriate logic.
    """
    return value.utcoffset() is not None


def make_aware(value: datetime, tz: tzinfo):
    if is_aware(value):
        raise ValueError("make_aware expects a naive datetime, got %s" % value)
        # This may be wrong around DST changes!
    return value.replace(tzinfo=tz)


def get_tzinfo(utc_offset=9):
    """
    timezone 정보 가져오기
    :param utc_offset: utc 기준시간. 기본값은 KST(+9)
    """
    return timezone(timedelta(hours=utc_offset))


def get_local_datetime(utc_offset=9):
    """
    Timezone이 지정된 datetime 현재시간
    :param utc_offset: utc 기준시간. 기본값은 KST(+9)
    """
    tz = get_tzinfo(utc_offset=utc_offset)
    return datetime.now().replace(tzinfo=tz)


def get_datetime_from_timestamp(second: str, aware=True, utc_offset=9):
    """
    String Type timestamp 를 형식에 맞는 Datetime 으로 변환\n
    :param second: Timestamp since the Epoch
    :param aware: Make aware 여부
    :param utc_offset: utc 기준시간. 기본값은 KST(+9)
    :return: Datetime
    """
    ts = float(second)

    if len(second) == 13:
        ts = ts / 1000

    dt = datetime.fromtimestamp(ts)

    if aware:
        tz = get_tzinfo(utc_offset=utc_offset)
        return make_aware(dt, tz)
    return dt


def get_datetime_from_iso_string(iso_string: str, aware=True, utc_offset=9):
    """
    ISO 형식의 문자열을 Datetime 으로 변환
    :param iso_string: ISO 형식의 문자열
    :param aware: Make aware 여부
    :param utc_offset: utc 기준시간. 기본값은 KST(+9)
    :return: Datetime 객체
    """
    dt = datetime.fromisoformat(iso_string)
    if aware and not dt.tzinfo:
        tz = get_tzinfo(utc_offset=utc_offset)
        return make_aware(dt, tz)
    return dt


def make_timestamp(dt: datetime, millisecond=True):
    """
    datetime 객체의 timestamp 변환
    :param dt: 대상
    :param millisecond: 밀리초 단위 변환 여부
    """
    timestamp = int(dt.timestamp())
    if millisecond:
        return timestamp * 1000
    return timestamp


def get_weekday_range(dt: datetime):
    """
    주어진 날짜의 주간 시작일과 종료일을 반환
    :param dt: 기준일
    :return: 시작일, 종료일 tuple
    """
    start = (dt - timedelta(days=dt.weekday())).replace(hour=0, minute=0, second=0)
    end = (start + timedelta(days=6)).replace(hour=23, minute=59, second=59)
    return start, end


def time_string_to_milliseconds(time_string):
    """
    "시:분:초" 형식의 문자열을 millisecond 단위의 정수로 변경하여 반환
    :param time_string: 시간 문자열
    :return: millisecond 정수
    """
    hours, minutes, seconds = time_string.split(":")
    total_milliseconds = (
        int(hours) * 3600000 + int(minutes) * 60000 + int(seconds) * 1000
    )
    return total_milliseconds



import os
import pickle
from json.decoder import JSONDecodeError
from typing import Any, Callable, Optional

from requests import Response, Session
from requests.adapters import HTTPAdapter
from requests.utils import cookiejar_from_dict, dict_from_cookiejar
from urllib3.util.retry import Retry

from ..exceptions import NotOkException


class AbstractRequestHelper:
    logger: Optional[Callable[[str, Any], Any]] = None

    def __init__(self):
        self._init_session()

    def _init_session(self):
        self.session = Session()
        self.session.mount(
            "https://",
            HTTPAdapter(
                max_retries=Retry(
                    total=3, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504]
                )
            ),
        )

    def set_logger(self, logger: Optional[Callable[[str, Any], Any]]):
        self.logger = logger

    @property
    def use_saved_session(self):
        return True

    @property
    def session_file_name(self):
        return os.path.join(
            os.path.split(__file__)[0],
            f"_session_{self.__class__.__name__}",
        )

    def _save_session_to_file(self):
        with open(self.session_file_name, "wb") as file:
            pickle.dump(dict_from_cookiejar(self.session.cookies), file)

    def _load_session_from_file(self):
        with open(self.session_file_name, "rb") as file:
            self.session.cookies = cookiejar_from_dict(pickle.load(file))

    def logging_response_error(self, msg: str, data=None):
        if self.logger:
            self.logger(msg, data)

    def get_response(self, res: Response, code=400, extra_ok=False):
        if not res.ok and not extra_ok:
            self.logging_response_error(msg=f"{code}")
            raise NotOkException(code=code)
        return res

    def get_response_json(
        self, res: Response, extra_ok=False, error_msg="응답이 올바르지 않습니다"
    ):
        try:
            data = res.json()
            ctx = {
                "status_code": res.status_code,
                **data,
            }
            if not res.ok and not extra_ok:
                self.logging_response_error(msg=error_msg, data=ctx)
                raise NotOkException(msg=error_msg, context=ctx)
            return data
        except JSONDecodeError:
            self.logging_response_error(
                msg="JSON Parse error", data={"status": res.status_code}
            )
            raise NotOkException(msg="데이터파싱에 실패했습니다")



from typing import Optional


class CodeMessageException(Exception):
    """
    내부 모듈에 사용
    """

    default_message = "모듈 에러가 발생했습니다"
    default_code = "_error"

    def __init__(
        self, code=None, msg: Optional[str] = None, context: Optional[dict] = None
    ):
        self.code = code or self.default_code
        self.msg: Optional[str] = msg or self.default_message
        self.context: Optional[dict] = context or {}
        message = f"code {code}.{msg}"
        super().__init__(message)

    def __str__(self):
        return "<{class_name} code: {code} message: {message}>".format(
            class_name=self.__class__.__name__,
            code=self.code,
            message=self.msg,
        )


class LoginException(CodeMessageException):
    default_message = "로그인 실패"
    default_code = "_login"


class NoAttrException(CodeMessageException):
    default_message = "필수 속성이 없습니다"
    default_code = "_noattr"


class NotOkException(CodeMessageException):
    default_message = "요청에 대한 응답이 올바르지 않습니다"
    default_code = "_request"
